#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <dlfcn.h>
#include <link.h>

#include "debug.h"

/* gcc -o -Wall -ldl debug.c -o debug */

// display level board with player position ('o' character).
void display(struct level_t * level, uint16_t pos) {
    int i;
    uint8_t cell_char;

    for (i=0; i < strlen(level->board); i++) {
        // get the current cell char.
        cell_char = level->board[i] ^ level->xor_mask;
        if ((i % level->board_width == 0) && (i != 0)) {  // add a line break at the end of each board lines.
            printf("\n");
        }
        if (i == pos) {  // if the current current cell is the starting one.
            c = 'o';
        }
        printf("%c ", c);  // print the current cell char.
    }
    printf("\n");
}

// load external functions from chall binary.
int load_funcs() {
    struct link_map * handle;
    void * bin_base;
    struct external_func * func;
    int i;

    handle = (struct link_map *) dlopen(CHALL_BINARY, RTLD_NOW);  // map the chall binary file.
    if (!handle) {
        fprintf(stderr, "dlopen() error!\n");
        return EXIT_FAILURE;
    }
    bin_base = (void *)handle->l_addr;  // get the chall binary base address.

    for (i=0; i < sizeof(funcs_nfo)/sizeof(funcs_nfo[0]); i++) {
        func = &funcs_nfo[i];  // get info about the external function (i.e. offset and name).
        *func->ptr = bin_base + func->offset;  // rebase and overwrite the function pointer.
    }

    return EXIT_SUCCESS;
}

int main(int argc, char * argv[]) {
    uint16_t current_pos;
    uint64_t moves;
    struct level_t * level;
    enum game_state res = PLAYING;
    int i;

    // load external functions (basically play and move).
    if (load_funcs() == 0) {
        printf("[+] Loaded functions:\n");
        for (i=0; i < sizeof(funcs_nfo)/sizeof(funcs_nfo[0]); i++) {
            printf(" %s: %p\n", funcs_nfo[i].name, *funcs_nfo[i].ptr);
        }
    } else {
        return EXIT_FAILURE;
    }

    // run level_x with arg_x as moves.
    for (i=1; i < argc; i++) {
        if (i <= sizeof(levels)/sizeof(levels[0])) {
            moves = strtoull(argv[i], NULL, 16);
            level = &levels[i-1];  // load level.
            current_pos = level->init_pos;

            if (moves != 0) {  // check if we are actually playing.
                res = play(&current_pos, level, moves);
            }

            printf("[+] Level %d: (moves = %llx)\n", i, moves);
            display(level, current_pos);
            switch (res) {
                case GAME_OVER:
                    printf("[-] Collision detected, game over!\n");
                    break;
                case SOLVED:
                    printf("[+] Level cleared!\n");
                    break;
            }
        } else {
            break;
        }
    }

    return EXIT_SUCCESS;
}